#!/usr/bin/perl
#
# Copyright (c) 2017  Peter Pentchev
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

use v5.010;
use strict;
use warnings;

use File::Basename qw(basename);
use File::Temp;
use Test::More;
use Test::Command;

my $prog = $ENV{TEST_PROG} // './stalepid';

my @usage_lines;

sub get_ok_output($ $) {
	my ($cmd, $desc) = @_;

	my $c = Test::Command->new(cmd => $cmd);
	$c->exit_is_num(0, "$desc succeeded");
	$c->stderr_is_eq('', "$desc did not output any errors");
	split /\n/, $c->stdout_value
}

sub get_error_output($ $) {
	my ($cmd, $desc) = @_;

	my $c = Test::Command->new(cmd => $cmd);
	$c->exit_isnt_num(0, "$desc failed");
	$c->stdout_is_eq('', "$desc did not output anything");
	split /\n/, $c->stderr_value
}

plan tests => 7;

# Get the usage lines.
subtest 'Usage output with -h' => sub {
	my $c = Test::Command->new(cmd => [$prog, '-h']);
	$c->exit_is_num(0, '-h succeeded');
	$c->stderr_is_eq('', '-h did not output any errors');
	my @lines = split /\n/, $c->stdout_value;
	BAIL_OUT('Too few lines in the -h output') unless @lines > 1;
	BAIL_OUT('Unexpected -h first line') unless $lines[0] =~ /^ Usage: \s+ stalepid /x;
	@usage_lines = @lines;
};

for my $num (0, 1, 3, 4) {
	subtest "Fail with $num arguments specified" => sub {
		my $c = Test::Command->new(cmd => [$prog, ($prog) x $num]);
		$c->exit_isnt_num(0, "failed with $num arguments");
		$c->stdout_is_eq('', "did not output anything with $num arguments");
		my @lines = split /\n/, $c->stderr_value;
		is_deeply \@lines, \@usage_lines, "output the usage message with $num arguments";
	};
}

my $d = File::Temp->newdir or BAIL_OUT 'Could not create a temporary directory';

# OK, let's build a string that is certainly longer than our process name.
subtest 'Delete a temporary file, another program running' => sub {
	my $long = join '..', $0, @ARGV;
	$long = $long x 2;

	my $fname = "$d/too-long.pid";
	open my $outf, '>', $fname or BAIL_OUT "Could not create $fname: $!";
	say $outf $$;
	close $outf or BAIL_OUT "Could not close $fname after writing: $!";

	ok -f $fname, "successfully created $fname";
	my @lines = get_ok_output([$prog, $fname, $long], 'delete, another program');
	is @lines, 0, 'delete, another program: did not output anything';
	ok ! -f $fname, "deleted $fname when another program was running";
};

subtest 'Leave it alone with the program still running' => sub {
	# OK, let's get our command name
	my $pid = $$; # do not let Perl's fork() fool us
	open my $ps, '-|', 'ps', '-c', '-o', 'command', '-p', $pid or BAIL_OUT "Could not fork for ps: $!";
	my $header = <$ps>;
	BAIL_OUT 'ps on ourselves did not output even a single line' unless defined $header;
	my $cmd = <$ps>;
	BAIL_OUT 'ps on ourselves did not output a second line' unless defined $cmd;
	chomp $cmd;
	diag "Got $cmd for our process's name";
	close $ps or BAIL_OUT 'ps on ourselves did not complete successfully';

	my $fname = "$d/ourselves.pid";
	open my $outf, '>', $fname or BAIL_OUT "Could not create $fname: $!";
	say $outf $$;
	close $outf or BAIL_OUT "Could not close $fname after writing: $!";

	ok -f $fname, "successfully created $fname";
	my @lines = get_ok_output([$prog, $fname, $cmd], "leave it alone: $cmd");
	is @lines, 0, "leave it alone: $cmd still running";
	ok -f $fname, "left $fname alone, $cmd still running";
};
